/* $Id: mga1064_main.c,v 1.9 1999/06/21 23:44:58 marcus Exp $
********************************************************************************
**      Matrox Mystique (mga1064sg) chipset driver
********************************************************************************
**
   Copyright (C)   1997            Matthias Grimrath
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
**      This code was derived from the following source of information:
** 
**      Matrox MGA-1064SG Developer Specification (Internet edition)
**      Matrox Millenium mga2064.c by Jonas Borgstrm
**      
*/

#include <kgi/maintainers.h>
#define MAINTAINER      Matthias_Grimrath
#define DRIVER_NAME     "Mystique chipset driver"
#define DRIVER_REV      "$Revision: 1.9 $"

#include <kgi/module.h>

#include <Matrox/mga_common_regs.h>  /* Register definitions */
#include <Matrox/mga_common_mmio.h>
#include <Matrox/mga_common_text16.h>
#include <Matrox/mga_common.h>
/* Driver specific things are included by mga_common_*.h */
#include <Matrox/mga1064_dac.h>

/* ??? configurable? */
#define HARDWARE_TEXT_CURSOR 0  /* Hardware cursor */

const char chipset_name_init[] = DRIVER_NAME " rev " DRIVER_REV " loaded";

/* Linearizable virtual buffer sizes */
const uint virtwidth[virtwidth_NUM] = { 
	512, 640, 768, 800, 832, 960, 1024, 1152, 1280, 1600, 1664, 1920, 2048
};

const struct kgi_chipset chipinfo = {
        "Matrox Graphics Inc",  /* Manufacturer         */
        "Mystique",             /* Model                */
        GM_TEXT16 | GM_1BIT |   /* Available modes      */
        GM_4BIT | GM_8BIT | 
        GM_15BIT | GM_16BIT | 
	GM_24BIT | GM_32BIT,
        0,                      /* Flags                */
        { 4095, 4095 },         /* Maximum resolution   */
	NULL,			/* Private data hook	*/
        0,                      /* to be set!           */
        { 6660000, 135000000 }  /* DCLK limits. Corrected in DAC code */
};

const struct mem_region mga_aperture = {
        MEM_NULL, 0, 0x4000, -1, "Mystique Aperture"
};

const struct mem_region mga_framebuffer = {
        MEM_NULL, 0, 0x800000, -1, "Mystique Frame Buffer"
};

static const struct mem_region mga_pseudodma = {
	MEM_NULL, 0, 0x800000, ~0x7FFFFF, "Mystique Pseudo-DMA"
};

const struct kgi_mmio_region graphic_mmio_linear_fb = {
        NULL,                   /* Next                  */
        "Mystique MMIO framebuffer",     /* Name         */
        MMAP_FRAMEBUFFER,       /* Type (Linear)        */
        MM_PROT_RW,             /* Protection map       */
     
        0x0,                    /* Total size of area   */
        0x0,                    /* Physical size        */
        0x0,                    /* bus address of the windows */
        0,                      /* Logical adress       */
     
        NULL,                   /* Set offset function  */
        NULL                    /* private              */
};

const struct kgi_mmio_region pingpong_mmio = {
	NULL,
	"Mystique ping-pong buffer",
	MMAP_TYPE_ACCEL,
	MM_PROT_RW,
	0,0,0,0,                /* size & address not needed here */
	NULL,
	NULL
};

const struct kgi_accel pingpong = {
	NULL,                   /* Next */
	NULL,                   /* display */
	"Mystique Ping-Pong",   /* Name */
	0,                      /* id */
	pingpong_exec,          /* feeds accelerator */
	pingpong_wait           /* wait for accelerator to finish */
};

/*      This table holds the default values for the VGA registers. A copy
**      of it will be filled in as appropriate during mode programming in
**      kgim_chipset_set_mode().
*/

static struct statetable default_state =
{
        0x00,                                           /* MISC */
        {
                0x03, 0x00, 0x00, 0x00, 0x00                    /* SEQ  */
        },
        {
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* CRT  */
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 0x1x */
                0x00
        },
        {
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* GRC  */
                0xFF
        },
        {
                0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, /* ATC  */
                0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
                0x00, 0x00, 0x0F, 0x00, 0x00
        },
        {
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00 /* ExtVga */
        }
};

static void restore_state(struct kgi_display *dpy, struct statetable *state)
{
	MISC_out8(dpy, state->MISC);
	restore_SEQ(dpy, state->SEQ);
	restore_CRT(dpy, state->CRT);
	restore_GRC(dpy, state->GRC);
	restore_ATC(dpy, state->ATC);
	restore_ExtVga(dpy, state->ExtVga);
}

/*
** Int handler.  Note that it must handle a shared IRQ line, so always
** check for the status bits!
*/
void mga_interrupt_handler(int irq, void *d, struct pt_regs *regs)
{
	struct kgi_display *dpy = (struct kgi_display *)d;
	GET_PRIVATE;
	kgi_uint status;
	
	status = mem_in32(STATUS);
	
	if (status & STAT_VSYNCPEN) {
		wake_up(&MGA_CP.vsync_wait);
	
		/* Enable next vsync interrupt */
		CRT_out8(dpy, CRT_in8(dpy, CRT_VSYNCEND) & 
			 ~CR11_CLEAR_VSYNC_IRQ, CRT_VSYNCEND);
		CRT_out8(dpy, CRT_in8(dpy, CRT_VSYNCEND) | 
			 CR11_CLEAR_VSYNC_IRQ, CRT_VSYNCEND);
	}
}

/*
** Called if the card is not initialized by the BIOS.  Must fail if it cannot
** cold boot the card.
*/

int mga_cold_init(struct kgi_display *dpy)
{
	GET_PRIVATE;
	int i;
	kgiu32 opt;
	kgi_uint sysm, sysn, sysp_s, sgram, refhcnt;

	if (private->cp.rev == 2) {
		sysm = 8;
		sysn = 112;
		sysp_s = (2 << 3) | 0;
		sgram = MGA_PCI_OPTION_HARDPWMSK;
		refhcnt = 0xF << 16;
	} else if (private->cp.rev == 3) {
		sysm = 8;
		sysn = 112;
		sysp_s = (2 << 3) | 0;
		sgram = MGA_PCI_OPTION_HARDPWMSK;
		refhcnt = 0x9 << 16;
	} else {
		NOTICE("Unsupported revision! Contact " MAINTAINER);
		return -ENODEV;
	}
	
	pcibios_read_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN, 
				  MGA_PCI_OPTION, &opt);

	/* Disable sync and data transfer for safety */
	ECRT_out8(dpy, CRTX_HSYNCOFF | CRTX_VSYNCOFF, CRTX_EXT1);
	SEQ_out8(dpy, SR01_DISPLAY_OFF, SEQ_CLOCK);
	
	EDAC_out8(dpy, 0, XVREFCTRL);
	udelay(100 * 1000);  /* Wait 100 milli seconds */
	
	/* Power up system PLL */
	opt |= MGA_PCI_OPTION_SYSPLLPDN | MGA_PCI_OPTION_SYSCLKDIS;
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				   MGA_PCI_OPTION, opt);
	EDAC_out8(dpy, sysm, XSYSPLLM);
	EDAC_out8(dpy, sysn, XSYSPLLN);
	EDAC_out8(dpy, sysp_s, XSYSPLLP);
	opt &= ~MGA_PCI_OPTION_SYSCLKDIS;
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				   MGA_PCI_OPTION, opt);
	while ((EDAC_in8(dpy, XSYSPLLSTAT) & PLL_LOCK) == 0) ;
	
	/* Power up pixel PLL */
	EDAC_out8(dpy, EDAC_in8(dpy, XPIXCLKCTRL) | PIX_PLL_POWER 
		  | PIX_CLK_DIS, XPIXCLKCTRL);
	EDAC_out8(dpy, 30, XPIXPLLCM);
	EDAC_out8(dpy, 108, XPIXPLLCN);
	EDAC_out8(dpy, 1, XPIXPLLCP);
	EDAC_out8(dpy, EDAC_in8(dpy, XPIXCLKCTRL) & ~PIX_CLK_DIS, XPIXCLKCTRL);
	while ((EDAC_in8(dpy, XPIXPLLSTAT) & PLL_LOCK) == 0) ;

	/* Power up LUT and DAC */
	EDAC_out8(dpy, EDAC_in8(dpy, XMISCCTRL) | RAMCS | DACPDN | VGA_8_DAC, 
		  XMISCCTRL);
	mem_out8(0, PALWTADD);
	for (i=0; i!=256; i++)
	        mem_out8(0, PALDATA);
	
	/* Enable system clock */
	opt |= MGA_PCI_OPTION_SYSCLKDIS;
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				   MGA_PCI_OPTION, opt);
	opt = (opt & ~MGA_PCI_OPTION_SYSCLKSEL_MASK) | 
	       MGA_PCI_OPTION_SYSCLKSEL_SYSTEM;
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				   MGA_PCI_OPTION, opt);
	opt &= ~MGA_PCI_OPTION_SYSCLKDIS;
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				   MGA_PCI_OPTION, opt);

	/* Enable pixel clock */
	EDAC_out8(dpy, EDAC_in8(dpy, XPIXCLKCTRL) | PIX_CLK_DIS, XPIXCLKCTRL);
	EDAC_out8(dpy, (EDAC_in8(dpy, XPIXCLKCTRL) & ~PIX_CLK_SEL_MASK) |
		  PIX_CLK_SEL_PIX, XPIXCLKCTRL);
	EDAC_out8(dpy, EDAC_in8(dpy, XPIXCLKCTRL) & ~PIX_CLK_DIS, XPIXCLKCTRL);
	
	/* Misc things */
	opt |= MGA_PCI_OPTION_FBMASKN_MASK | sgram;
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				   MGA_PCI_OPTION, opt);
	EDAC_out8(dpy, XGEN_IOGSYNCDIS, XGENCTRL);
	
	/* Power up video RAM */
	mem_out32(CASLTNCY | RCDELAY | RASMIN_MASK, MCTLWTST);
	udelay(200);
	mem_out32(MEMRESET, MACCESS);
	udelay(100);
	mem_out32(MEMRESET | JEDECRST, MACCESS);
	opt = (opt & ~MGA_PCI_OPTION_RFHCNT) | refhcnt;
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				   MGA_PCI_OPTION, opt);

	return -EOK;
}

/*
** Called if the card was coldbooted and the driver is to be removed.  It is
** supposed to turn off the card as much as possible including disabling
** device response to IO/Mem regions.
*/

void mga_cold_shutdown(struct kgi_display *dpy)
{
	GET_PRIVATE;
	kgiu32 com;
	
	/* !!! By far not enough */
	
	pcibios_read_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN, 
				  PCI_COMMAND, &com);
	com = com & ~(PCI_COMMAND_IO | PCI_COMMAND_MEMORY);
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				   PCI_COMMAND, com);
}

/*
**      KGI module interface
*/

/* Mode specific frame buffers */
void adjust_timing(struct kgi_mode *mode)
{
        int foo = (mode->request.graphtype == KGIGT_TEXT16) ? 
	  mode->request.dpp.x : 8;

        #define ADJUST(X) mode->x.X -= (mode->x.X % foo)

        ADJUST(width);
        ADJUST(blankstart);
        ADJUST(syncstart);
        ADJUST(syncend);
        ADJUST(blankend);

        if (mode->x.total % foo) {
                mode->x.total += foo - (mode->x.total % foo);
        }

        if (((mode->x.syncend / foo) & 0x3F) ==
	    ((mode->x.syncstart / foo) & 0x3F)) {
                mode->x.syncend -= foo;
        }

        if (((mode->x.blankend / foo) & 0x7F) ==
	    ((mode->x.blankstart / foo) & 0x7F)) {
                mode->x.blankend -= foo;
        }

        if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) {
                mode->y.syncend--;
        }

        if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) {
                mode->y.blankend--;
        }

        #undef ADJUST
}

int adjust_virtual(struct kgi_mode *mode)
{
	int i,j;
#if 1
	/* Block mode (= 100% faster than "normal" mode) requires this */
        switch (mode->request.graphtype) {
        case KGIGT_8BIT: 
        case KGIGT_24BIT: j = 63; break;
	case KGIGT_15BIT:
        case KGIGT_16BIT:
        case KGIGT_32BIT: j = 31; break;         
        default: INTERNAL_ERROR; return -EINVAL;
        }

	/* Find smallest linearizable virtual size */
	for (i=0; (i!=virtwidth_NUM) && 
	          ((virtwidth[i] < mode->request.virt.x) ||
		   ((virtwidth[i] & j) != 0)); i++) ;
	if (i==virtwidth_NUM)
	        return -EINVAL;
	mode->request.virt.x = virtwidth[i];
#else
	/* for tight vidmem disable accel and activate this */
        switch (mode->request.graphtype) {
        case KGIGT_8BIT:  i = 15; break;         
	case KGIGT_15BIT:
        case KGIGT_16BIT: i = 7;  break;         
        case KGIGT_24BIT: i = 15; break;
        case KGIGT_32BIT: i = 3;  break;         
        default: INTERNAL_ERROR;   return;
        }

	if ((x=mode->request.virt.x & i)!=0) {
		mode->request.virt.x = mode->request.virt.x+round-x;
	}
	/*mode->request.virt.y not necessary*/
#endif	
	return 0;
}

int kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
	int i;
	kgiu32 option;
	struct chipset_private *private;
	
	if ((i = mga_chipset_init(card, dpy, sizeof(struct chipset_private),
				  PCI_DEVICE_ID_MATROX_MYS))!=EOK) {
		mga_chipset_done(dpy);
		return i;
	}

	private = KGIM_PRIVATE(dpy, chipset);
	
	private->pseudo_dma = mga_pseudodma;
	pcibios_read_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN,
				  PCI_BASE_ADDRESS_2, 
				  &private->pseudo_dma.base);
	private->pseudo_dma.base &= 0xFF800000;
	if (mem_check_region(&private->pseudo_dma)) {
                ERROR("  PSEUDODMA busy.");
		mga_chipset_done(dpy);
                return -E(CHIPSET, UNKNOWN);
        }
        mem_claim_region(&private->pseudo_dma);

	pcibios_read_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN, 
				  MGA_PCI_OPTION, &option);
	private->info.blkmode = (option & MGA_PCI_OPTION_HARDPWMSK) ?
  	                        ATYPE_BLK : ATYPE_RPL;

	private->pingpong_mmio = pingpong_mmio;
	MGA_CP.graphic_mmio_linear_fb.next = &private->pingpong_mmio;
	
	private->pingpong = pingpong;
	private->pingpong.dpy = dpy;
	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	GET_PRIVATE;

        if (private->pseudo_dma.baseptr)
                mem_free_region(&private->pseudo_dma);

	mga_chipset_done(dpy);

        NOTICE("%s", DRIVER_NAME " rev " DRIVER_REV " removed");
}

static void mga_set_origin(struct kgi_display *dpy, kgi_uint x, uint y)
{
	IS_CRITICAL;
	uint offset;

	switch (dpy->mode.request.graphtype) {

	case KGIGT_1BIT:
	case KGIGT_4BIT:
	case KGIGT_8BIT:
		offset = (y*dpy->mode.request.virt.x + x) / 8;
		break;
		
	case KGIGT_24BIT:
		offset = ((y*dpy->mode.request.virt.x + x) / 8) * 3;
		break;

	case KGIGT_15BIT:
	case KGIGT_16BIT:
		offset = (y*dpy->mode.request.virt.x + x) / 4;
		break;

	case KGIGT_32BIT:
		offset = (y*dpy->mode.request.virt.x + x) / 2;
		break;

	default:
		INTERNAL_ERROR;
		return;
	}

	DEBUG2("x %d y %d offset %d",x,y,offset);
		
	ENTER_CRITICAL;

	CRT_out8(dpy, offset, CRT_STARTADDR_L);
	CRT_out8(dpy, offset >> 8, CRT_STARTADDR_H);
	ECRT_out8(dpy, (ECRT_in8(dpy, CRTX_EXT0) & ~CRTX_STARTADD_MASK) | 
		  ((offset >> 16) & CRTX_STARTADD_MASK), CRTX_EXT0);

	/* Can't do dots.  Or is there a way? */
	LEAVE_CRITICAL;	
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
        IS_CRITICAL;
   
        GET_PRIVATE;
        struct statetable setup = default_state;
        int DCLKdiv = 8;
	int bpptype;
	int bpp;
        int textx,texty;
        int wd, /* logical width of a framebuffer line (stride ?) */
   	    hd, /* hor display end */
	    hs, /* hor sync start */
            he, /* hor sync end */
            hb, /* hor blank start */
            hc, /* hor blank end */
            ht, /* hor total */
            vd, /* vert display end */
            vs, /* vert sync start */
            ve, /* vert sync end */
            vb, /* vert blank start */
            vc, /* vert blank end */
            vt, /* vert total */
            ln; /* line compare */

	DEBUG2("Setting mode... (%dx%d %d)\n", mode->request.virt.x,
                mode->request.virt.y, mode->request.graphtype);

        textx = mode->request.dpp.x;
        texty = mode->request.dpp.y;
   
        /*      Set up bits per pixel and logical screen width
        */
        switch (mode->request.graphtype) {

        case KGIGT_TEXT16:
        case KGIGT_8BIT:
	case KGIGT_15BIT:
        case KGIGT_16BIT:
        case KGIGT_24BIT:
        case KGIGT_32BIT:
                break;

        default:
                INTERNAL_ERROR;
                return;
        }

        if (mode->request.graphtype == KGIGT_TEXT16) {

		bpp     = 16;
		bpptype = MAP_16_BPP;
                wd = mode->request.virt.x * 2 / 4;
                /* Due to the addressing mode used */

                setup.CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC |
		        CR17_CGA_BANKING | CR17_HGC_BANKING;
      
                setup.CRT[0x14] |= CR14_UNDERLINE_MASK;

                setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;
      
                setup.GRC[0x05] |= GR5_ODD_EVEN;
                setup.GRC[0x06] |= GR6_CHAIN_OE;

                setup.GRC[0x06] |= GR6_MAP_A0_64;

                switch ((DCLKdiv = textx)) {

                case 8:
                        setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
                        break;

                case 9:
                        setup.ATC[0x10] |= AR10_9DOT_CHARS;
                        break;

                default:
                        INTERNAL_ERROR;
                        return;
                }

                setup.ATC[0x13] = (textx == 9) ? 0x08 : 0x00;

        } else { 

                setup.SEQ[0x00] = 0x00;
                setup.SEQ[0x01] = 0x01; /*0x09*/
                setup.SEQ[0x02] = 0x0F;
                setup.SEQ[0x03] = 0x00;
                setup.SEQ[0x04] = 0x0E;

                setup.GRC[0x00] = 0x00;
                setup.GRC[0x01] = 0x00;
                setup.GRC[0x02] = 0x00;
                setup.GRC[0x03] = 0x00;
                setup.GRC[0x04] = 0x00;
                setup.GRC[0x05] = 0x40;
                setup.GRC[0x06] = 0x05;
                setup.GRC[0x07] = 0x0F;
                setup.GRC[0x08] = 0xFF;

                setup.ATC[0x10] = 0x41;
                setup.ATC[0x11] = 0xFF;
                setup.ATC[0x12] = 0x0F;
                setup.ATC[0x13] = 0x00;
                setup.ATC[0x14] = 0x00;

                setup.CRT[0x17] = 0xC3;

                switch(mode->request.graphtype) {

                case KGIGT_8BIT:
                        bpp     = 8;
			bpptype = MAP_8_BPP;
                        break;

		case KGIGT_15BIT:
		case KGIGT_16BIT:
                        bpp     = 16;
			bpptype = MAP_16_BPP;
                        break;

                case KGIGT_24BIT:
                        bpp     = 24;
			bpptype = 0; /* Non standard */
                        break;

                case KGIGT_32BIT:
                        bpp     = 32;
			bpptype = MAP_32_BPP;
                        break;

                default:
                        INTERNAL_ERROR;
                        return;
                }

		wd    = mode->request.virt.x * bpp / 128;
                texty = mode->magnify.y;
        }
        
        /*
        **      Set up timing
        */
   
        hd = (mode->x.width / DCLKdiv) - 1;
        hs = (mode->x.syncstart / DCLKdiv);
        he = (mode->x.syncend / DCLKdiv);
        ht = (mode->x.total / DCLKdiv) - 5;

        vd = mode->y.width - 1;
        vs = mode->y.syncstart;
        ve = mode->y.syncend;
        vt = mode->y.total - 2;

        ln = 0x7ff;

	setup.MISC = ((mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC) |
		((mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC) |
		MISC_CLOCK_1X | MISC_COLOR_IO;
   
        if (mode->request.graphtype != KGIGT_TEXT16) {
                /* No Overscan in MGA modes */
                hb = (mode->x.width / DCLKdiv) - 1;
                hc = (mode->x.total / DCLKdiv) - 1;
                vb = mode->y.width - 1;
                vc = mode->y.total - 1;

                setup.ExtVga[5] = 0;
        
                setup.ExtVga[0] = (wd & 0x300) >> 4;
                setup.ExtVga[1] = ((ht & 0x100) >> 8) | ((hb & 0x100) >> 7)
                        | ((hs & 0x100) >> 6) | (hc & 0x40);

                setup.ExtVga[2] = ((vt & 0xC00) >> 10) |
                        ((vd & 0x400) >> 8) | 
                        ((vb & 0xC00) >> 7) |
                        ((vs & 0xC00) >> 5) |
                        ((ln & 0x400) >> 3);

		setup.ExtVga[3] = ((bpp / 8) - 1) | CRTX_MGAMODE;
                setup.ExtVga[4] = 0;

        } else {
                /* Use overscan values from KGI Manager */
                hb = mode->x.blankstart / DCLKdiv;
                hc = mode->x.blankend / DCLKdiv;
                vb = mode->y.blankstart - 1;
                vc = mode->y.blankend + 1;
        }

        setup.CRT[0]  = ht;
        setup.CRT[1]  = hd;
        setup.CRT[2]  = hb & 0xFF;
        setup.CRT[3]  = (hc & 0x1F) | CR03_IS_VGA;
        setup.CRT[4]  = hs;
        setup.CRT[5]  = ((hc & 0x20) << 2) | (he & 0x1F);
        setup.CRT[6]  = vt & 0xFF;
        setup.CRT[7]  = ((vt & 0x100) >> 8 ) | ((vd & 0x100) >> 7 ) |
                ((vs & 0x100) >> 6 ) | ((vb & 0x100) >> 5 ) |
                ((ln & 0x100) >> 4) | ((vt & 0x200) >> 4 ) | 
                ((vd & 0x200) >> 3 ) | ((vs & 0x200) >> 2 );
        setup.CRT[8]  = 0;
        setup.CRT[9]  = ((ln & 0x200) >> 3) | ((vb & 0x200) >> 4) | 
                ((texty - 1) & 0x1F);
        setup.CRT[16] = vs & 0xFF;
        setup.CRT[17] = (ve & 0x0F) | CR11_DISABLE_VSYNC_IRQ; /* clear too */
        setup.CRT[18] = vd & 0xFF;
        setup.CRT[19] = wd & 0xFF;
        setup.CRT[21] = vd & 0xFF;
        setup.CRT[22] = vc & 0xFF;
	setup.CRT[24] = ln & 0xFF;
	
        /*      Ok.  We'll do the actual hardware programming now.
        */

	/* Wait for vsync except if it's the first mode set */
	/* !!! Causes reentrant problems, so left out first */
	
        ENTER_CRITICAL;
	ECRT_out8(dpy, ECRT_in8(dpy, CRTX_EXT1) | CRTX_HSYNCOFF |
		  CRTX_VSYNCOFF, CRTX_EXT1); /* turn off sync pulses */
        kgim_monitor_set_mode(dpy, mode);
        kgim_ramdac_set_mode(dpy, mode);
        kgim_clock_set_mode(dpy, mode);
        kgim_accel_set_mode(dpy, mode);
        restore_state(dpy, &setup); /* also enables sync signals */
	CRT_out8(dpy, (CRT_in8(dpy, CRT_VSYNCEND) & ~CR11_DISABLE_VSYNC_IRQ) |
		 CR11_CLEAR_VSYNC_IRQ, CRT_VSYNCEND);
        LEAVE_CRITICAL;

	memset(MGA_FRAMEBUFFER.baseptr, 0, MGA_CP.chipinfo.ramsize);
        memcpy(&(dpy->mode), mode, sizeof(dpy->mode));

        NOTICE2("setting up display...");

        private->cp.split = -1;
        private->cp.org_xdot = private->cp.org_ydot = private->cp.org_offs= -1;
	private->cp.bpp = bpp;

	dpy->fb.bpp = bpptype;
        if (mode->request.graphtype == KGIGT_TEXT16) {
                dpy->set_font   = mga_text_set_font;
                dpy->set_split  = mga_set_split;
                dpy->set_frame  = mga_set_frame;
                dpy->set_origin = mga_text_set_origin;

                /*      provide hardware cursor in text mode
                */
                if (HARDWARE_TEXT_CURSOR) {
                        dpy->show_cursor = mga_show_hw_text_cursor;
                        dpy->hide_cursor = mga_hide_hw_text_cursor;
                        dpy->undo_cursor = NULL;
                } else {
                        dpy->show_cursor = mga_show_sw_text_cursor;
                        dpy->hide_cursor = mga_hide_sw_text_cursor;
                        dpy->undo_cursor = mga_undo_sw_text_cursor;
                }

                /*      provide software cursor too in case DAC didn't already.
                */
                if (! dpy->show_pointer) {
                        dpy->show_pointer = mga_show_sw_text_pointer;
                        dpy->hide_pointer = mga_hide_sw_text_pointer;
                        dpy->undo_pointer = mga_undo_sw_text_pointer;
                }

                dpy->fb.size.x = mode->request.virt.x;
                dpy->fb.size.y = mode->request.virt.y;
		dpy->fb.bpp    = MAP_16_BPP;
                dpy->fb.pbuf   = MGA_FRAMEBUFFER.baseptr;
                dpy->fb.zbuf   = NULL;
                dpy->fb.ops    = &mgatext_color_io;
		
        } else {

                dpy->fb.size.x  = mode->request.virt.x;
                dpy->fb.size.y  = mode->request.virt.y;
                dpy->fb.pbuf    = NULL;
                dpy->fb.zbuf    = NULL;
                dpy->fb.ops     = NULL;

                dpy->show_pointer       = NULL;
                dpy->hide_pointer       = NULL;
                dpy->undo_pointer       = NULL;

                dpy->show_cursor        = NULL;
                dpy->hide_cursor        = NULL;
                dpy->undo_cursor        = NULL;

                dpy->set_font           = NULL;
                dpy->set_split          = mga_set_split;
                dpy->set_frame          = mga_set_frame;
                dpy->set_origin         = mga_set_origin;
        }
	private->cp.fb_size = dpy->fb.size.x * dpy->fb.size.y;
	
        NOTICE1("Mode set complete");
}

int kgim_chipset_command(struct kgi_graphic *state, uint cmd, uint arg)
{
	struct kgi_device *dev = &(state->dev);
	struct chipset_private *private = KGIM_PRIVATE(dev->dpy, chipset);
	
	switch(cmd) {
		
	case CHIP_SETVISFRAME:
		DEBUG2("x %d y %d", ((kgi_coord *)(state->io_buf))->x, 
		       ((kgi_coord *)(state->io_buf))->y);
		kgi_set_origin(dev, dev->frame, 
			       ((kgi_coord *)(state->io_buf))->x,
			       ((kgi_coord *)(state->io_buf))->y);
		return EOK;
	
	case CHIP_SETSPLITLINE: /* !!! Fail in SVGA mode */
		if (arg > dev->mode.virt.y)
		        return -E(CHIPSET,BOGUS);
		kgi_set_split(dev, arg);
		return EOK;

	case CHIP_MYST_GETRAYPOS:
		*((kgis32 *)(state->io_buf)) = mem_in32(VCOUNT);
		state->io_res = state->io_buf;
		return EOK;
		
	case CHIP_MYST_WAITRAYPOS:
		sleep_on(&MGA_CP.vsync_wait);
		return EOK;

		/* Returns card information to userspace */
	case CHIP_MYST_GETINFO:
		state->io_res = &private->info;
		return EOK;

	default:
		return -E(CHIPSET,UNKNOWN);
	}
}
